package gs.uuid.v0

import com.fasterxml.uuid.Generators

/** Alias for the `java.util.UUID` type, which represents a 128-bit (16-byte)
  * value.
  *
  * ## ID Generation
  *
  * This library provides generator implementations for the following types of
  * UUID:
  *
  *   - Type 4
  *   - Type 7
  *
  * These implementations are provided by JUG.
  *
  * ## Serialization
  *
  * This library supports the following representations:
  *
  *   - Hexadecimal string without dashes
  *   - Byte array
  *
  * Serialization code is based on Jackson Databind (Apache 2.0).
  */
opaque type UUID = java.util.UUID

object UUID:
  /** Express any `java.util.UUID` as a GS UUID.
    *
    * @param uuid
    *   The input UUID.
    * @return
    *   The aliased value.
    */
  def apply(uuid: java.util.UUID): UUID = uuid

  /** Parse a byte array as a [[UUID]].
    *
    * See `toBytes()` for the inverse of this function.
    *
    * @param bytes
    *   The array of bytes to parse.
    * @return
    *   The new [[UUID]], or `None` if the bytes do not represent a valid UUID.
    */
  def fromBytes(bytes: Array[Byte]): Option[UUID] =
    scala.util
      .Try(UUIDFormat.fromBytes(bytes))
      .map(uuid => Some(apply(uuid)))
      .getOrElse(None)

  given CanEqual[UUID, UUID] = CanEqual.derived

  /** Generate a new UUID.
    *
    * @param G
    *   The [[Generator]] type class instance.
    * @return
    *   The new UUID.
    */
  def generate(
  )(
    using
    G: Generator
  ): UUID = G.next()

  /** @return
    *   New v4 UUID (Random).
    */
  def v4(): UUID = Generator.version4.next()

  /** @return
    *   New v7 UUID (Epoch + Counter + Random).
    */
  def v7(): UUID = Generator.version7.next()

  /** Parse the given string as a UUID.
    *
    * @param str
    *   The UUID, which is expected to be in a hexadecimal format with no
    *   dashes.
    * @return
    *   The parsed UUID value, or `None` if the value does not represent a UUID.
    */
  def parse(str: String): Option[UUID] = fromString(str)

  /** Parse the given string as a UUID.
    *
    * @param str
    *   The UUID, which is expected to be in a hexadecimal format with no
    *   dashes.
    * @return
    *   The parsed UUID value, or `None` if the value does not represent a UUID.
    */
  def fromString(str: String): Option[UUID] =
    scala.util
      .Try(UUIDFormat.fromHex(str))
      .map(uuid => Some(apply(uuid)))
      .getOrElse(None)

  extension (uid: UUID)
    /** @return
      *   The underlying `java.util.UUID`.
      */
    def toUUID(): java.util.UUID = uid

    /** @return
      *   The byte array representation of this UUID.
      */
    def toBytes(): Array[Byte] = UUIDFormat.toBytes(uid)

    /** Serialize this UUID.
      *
      * @param dashes
      *   Whether to use dashes in the output (default = false).
      * @return
      *   Hexadecimal string representation of this UUID.
      */
    def str(dashes: Boolean = false): String = withoutDashes()

    /** @return
      *   Hexadecimal string representation of this UUID, without dashes.
      */
    def withoutDashes(): String = UUIDFormat.toHexWithoutDashes(uid)

    /** @return
      *   Hexadecimal string representation of this UUID, with dashes.
      */
    def withDashes(): String = UUIDFormat.toHexWithDashes(uid)

    /** @return
      *   The least significant bits of this UUID.
      */
    def lsb(): Long = uid.getLeastSignificantBits()

    /** @return
      *   The most significant bits of this UUID.
      */
    def msb(): Long = uid.getMostSignificantBits()

    /** @return
      *   `true` if this UUID is `0`, `false` otherwise.
      */
    def isZero(): Boolean = lsb() == 0L && msb() == 0L

  /** Type class for UUID generation.
    */
  trait Generator:
    /** Generate a new UUID.
      */
    def next(): UUID

  object Generator:
    /** Instantiate a new Type 4 generator.
      */
    lazy val version4: Generator = new Version4

    /** Instantiate a new Type 7 generator.
      */
    lazy val version7: Generator = new Version7

    /** Type 4 (Random) implementation of a UUID generator.
      */
    final class Version4 extends Generator:
      private val gen           = Generators.randomBasedGenerator()
      override def next(): UUID = gen.generate()

    /** Type 7 (Unix Epoch Time + Counter + Random) implementation of a UUID
      * generator. Consider using this rather than Type 1 or Type 6.
      *
      * This type is defined in [IETF New UUID Formats
      * Draft](https://www.ietf.org/archive/id/draft-peabody-dispatch-new-uuid-format-04.html#name-uuid-version-7)
      */
    final class Version7 extends Generator:
      private val gen           = Generators.timeBasedEpochGenerator()
      override def next(): UUID = gen.generate()
